﻿package org.papervision3d.objects.primitives {
	import org.papervision3d.Papervision3D;
	import org.papervision3d.core.geom.*;
	import org.papervision3d.core.geom.renderables.Triangle3D;
	import org.papervision3d.core.geom.renderables.Vertex3D;
	import org.papervision3d.core.math.NumberUV;
	import org.papervision3d.core.proto.*;	
	import org.papervision3d.objects.*;

	/**
	* The Plane class lets you create and display flat rectangle objects.
	* <p/>
	* The rectangle can be divided in smaller segments. This is usually done to reduce linear mapping artifacts.
	* <p/>
	* Dividing the plane in the direction of the perspective or vanishing point, helps to reduce this problem. Perspective distortion dissapears when the plane is facing straignt to the camera, i.e. it is perpendicular with the vanishing point of the scene.
	*/
	public class PlanePmem extends TriangleMesh3D
	{
		
		public var point_list:Array=[];
		public var point_list2:Array=[];
		public var point_mcs:Array=[];
		public var close_point_mcs:Array=[];
		public var point_exe_flgs:Array=[];
		public var pointer:DisplayObject3D;
		public var pointw:Number;
		public var pointh:Number;
		public var point_vector_mcs:Array=[];
		public var point_gravity_mcs:Array=[];
		
		public var wind_point_status:Array=[];
		
		public var pair_plane:PlanePmem=null;
		
		/**
		* Number of segments horizontally. Defaults to 1.
		*/
		public var segmentsW :Number;
	
		/**
		* Number of segments vertically. Defaults to 1.
		*/
		public var segmentsH :Number;
	
		/**
		* Default size of Plane if not texture is defined.
		*/
		static public var DEFAULT_SIZE :Number = 500;
	
		/**
		* Default size of Plane if not texture is defined.
		*/
		static public var DEFAULT_SCALE :Number = 1;
	
		/**
		* Default value of gridX if not defined. The default value of gridY is gridX.
		*/
		static public var DEFAULT_SEGMENTS :Number = 1;
	
	
		// ___________________________________________________________________________________________________
		//                                                                                               N E W
		// NN  NN EEEEEE WW    WW
		// NNN NN EE     WW WW WW
		// NNNNNN EEEE   WWWWWWWW
		// NN NNN EE     WWW  WWW
		// NN  NN EEEEEE WW    WW
	
		/**
		* Create a new Plane object.
		* <p/>
		* @param	material	A MaterialObject3D object that contains the material properties of the object.
		* <p/>
		* @param	width		[optional] - Desired width or scaling factor if there's bitmap texture in material and no height is supplied.
		* <p/>
		* @param	height		[optional] - Desired height.
		* <p/>
		* @param	segmentsW	[optional] - Number of segments horizontally. Defaults to 1.
		* <p/>
		* @param	segmentsH	[optional] - Number of segments vertically. Defaults to segmentsW.
		* <p/>
		*/
		public function PlanePmem( material:MaterialObject3D=null, width:Number=0, height:Number=0, segmentsW:Number=0, segmentsH:Number=0 )
		{
			super( material, new Array(), new Array(), null );
	
			this.segmentsW = segmentsW || DEFAULT_SEGMENTS; // Defaults to 1
			this.segmentsH = segmentsH || this.segmentsW;   // Defaults to segmentsW
	
			var scale :Number = DEFAULT_SCALE;
	
			if( ! height )
			{
				if( width )
					scale = width;
	
				if( material && material.bitmap )
				{
					width  = material.bitmap.width  * scale;
					height = material.bitmap.height * scale;
				}
				else
				{
					width  = DEFAULT_SIZE * scale;
					height = DEFAULT_SIZE * scale;
				}
			}
	
			buildPlane( width, height );
			
			pointer=new DisplayObject3D;
			pointer.autoCalcScreenCoords=true;
			addChild(pointer);
			pointw=width/segmentsW;
			pointh=height/segmentsH;
			
			var a:int=0;
			var mc;
			for(a=0;a<point_mcs.length;a++){
				mc=point_mcs[a];
				wind_point_status[a]={defx:mc.x, defy:mc.y, defz:mc.z, nowx:mc.x, nowy:mc.y, nowz:mc.z, brad:0, brx:0, bry:0, brz:0, basebrx:0, basebry:0, basebrz:0, vector_flg:false, limit:200};
			}
		}
		
		public function set_point_direct2mz():void{
			var a:int=0;
			var b:int=0;
			var prev;
			var gap:Number=0;
			for(a=0;a<point_list2.length;a++){
				for(b=0;b<point_list2[a].length;b++){
					point_list2[a][b].x=0;
					point_list2[a][b].y=pointh*0.5+pointh*a;
					point_list2[a][b].z=-pointw*b;
				}
			}
			set_pair_pos();
		}
		
		
		public function set_point_limit_pos():void{
			var a:int=0;
			var b:int=0;
			var prev;
			var gap:Number=0;
			for(a=0;a<point_list2.length;a++){
				for(b=1;b<point_list2[a].length;b++){
					prev=point_list2[a][b-1];
					gap=get_gap(prev, point_list2[a][b]);
					if(gap>pointw){
						pv_set_limit_ball(prev, point_list2[a][b], pointw);
						//set_limit_pos(prev, point_list2[a][b], pointw);
					}
				}
			}
			set_pair_pos();
		}
		
		public function set_point_limit_pos2():void{
			var a:int=0;
			var b:int=0;
			var prev;
			var gap:Number=0;
			var wx:Number=0;
			var wy:Number=0;
			var wz:Number=0;
			var tx:Number=0;
			var ty:Number=0;
			var tz:Number=0;
			var sp:Number=0.5;
			for(a=0;a<point_list2.length;a++){
				for(b=1;b<point_list2[a].length;b++){
					
					prev=point_list2[a][b-1];
					
					/*
					wx=point_list2[a][b].x;
					wy=point_list2[a][b].y;
					wz=point_list2[a][b].z;
					set_limit_pos(prev, point_list2[a][b], pointw);
					tx=point_list2[a][b].x;
					ty=point_list2[a][b].y;
					tz=point_list2[a][b].z;
					point_list2[a][b].x=wx;
					point_list2[a][b].y=wy;
					point_list2[a][b].z=wz;
					*/
					tx=prev.x;
					ty=prev.y;
					tz=prev.z;
					point_list2[a][b].x+=(tx-point_list2[a][b].x)*sp;
					point_list2[a][b].y+=(ty-point_list2[a][b].y)*sp;
					point_list2[a][b].z+=(tz-point_list2[a][b].z)*sp;
					
					//point_list2[a][b].x=Math.round(point_list2[a][b].x);
					//point_list2[a][b].y=Math.round(point_list2[a][b].y);
					//point_list2[a][b].z=Math.round(point_list2[a][b].z);
				}
			}
			
			set_pair_pos();
		}
		
		public function set_pair_pos():void{
			var a:int=0;
			var b:int=0;
			if(pair_plane!=null){
				var model;
				var target;
				for(a=0;a<pair_plane.point_list2.length;a++){
					for(b=0;b<pair_plane.point_list2[a].length;b++){
						model=point_list2[a][b];
						target=pair_plane.point_list2[a][b];
						target.x=model.x;
						target.y=model.y;
						target.z=model.z;
					}
				}
			}
		}
		
		public function set_base_point2bar(bar):void{
			var p0=bar.extra.p0;
			var p1=bar.extra.p1;
			var mc0=point_list2[0][0];
			var mc1=point_list2[1][0];
			mc0.x=p0.sceneX;
			mc0.y=p0.sceneY;
			mc0.z=p0.sceneZ;
			mc1.x=p1.sceneX;
			mc1.y=p1.sceneY;
			mc1.z=p1.sceneZ;
		}
		
		public function set_limit_pos(base_point, target_point, radius:Number):void{
			var baserad:Number=270*get_k2r();
			var baseradh:Number=0*get_k2r();
			var camrad:Number=0;
			var camradh:Number=0;
			
			camrad=get_angle(Math.round(base_point.x), Math.round(base_point.z), Math.round(target_point.x), Math.round(target_point.z));
			target_point.x=Math.cos(camrad)*radius+base_point.x;
			target_point.z=Math.sin(camrad)*radius+base_point.z;
			
			var vz:Number=target_point.z;
			//camradh=get_angle(base_point.x, base_point.y, target_point.x, target_point.y);
			camradh=get_angle(Math.round(base_point.x), Math.round(base_point.y), Math.round(target_point.x), Math.round(target_point.y));
			
			target_point.z=Math.cos(camradh)*vz+base_point.z;
			target_point.y=Math.sin(camradh)*radius+base_point.y;
		}
		
		public function addWind(no:int, vecx:Number=0, vecy:Number=0, vecz:Number=0, flg:Boolean=false):void{
			var target:Object=wind_point_status[no];
			
			var ratios:Object={x:0, y:0, z:0};
			var runit:Number=0;
			var vec:Number=Math.abs(vecx)+Math.abs(vecy)+Math.abs(vecz);
			if(vec>0){
				runit=1/vec;
				ratios.x=runit*Math.abs(vecx)*0.9;
				ratios.y=runit*Math.abs(vecy)*0.9;
				ratios.z=runit*Math.abs(vecz)*0.9;
			}
			target.basebrx+=Math.sin(target.brad)*(target.brx);
			target.basebry+=Math.sin(target.brad)*(target.bry);
			target.basebrz+=Math.sin(target.brad)*(target.brz);
			
			target.brx=vecx*8;
			target.bry=vecy*8;
			target.brz=vecz*8;
			
			var num:Number=0;
			var limit:Number=10;
			
			num=target.brx;
			limit=target.limit*ratios.x-Math.abs(target.basebrx);
			if(Math.abs(num)>limit)num=limit*((num>0)?1:-1);
			target.brx=num;
			
			num=target.bry;
			limit=target.limit*ratios.y-Math.abs(target.basebry);
			if(Math.abs(num)>limit)num=limit*((num>0)?1:-1);
			target.bry=num;
			
			num=target.brz;
			limit=target.limit*ratios.z-Math.abs(target.basebrz);
			if(Math.abs(num)>limit)num=limit*((num>0)?1:-1);
			target.brz=num;
			
			
			target.vector_flg=true;
			target.brad=0;
			
			if(!flg)return;
			
			var gap:Number=0;
			var a:int=0;
			var b:Number=0;
			var c:Number=0;
			for(a=0;a<close_point_mcs[no].length;a++){
				b=get_gap(point_mcs[no], close_point_mcs[no][a]);
				if(b>gap)gap=b;
			}
			gap*=1;
			for(a=0;a<close_point_mcs[no].length;a++){
				b=get_gap(point_mcs[no], close_point_mcs[no][a]);
				c=(gap-b)/gap;
				addWind(close_point_mcs[no][a].extra.no, vecx*c, vecy*c, vecz*c, false);
			}
		}
		
		public function oef_wind_motion():void{
			var a:int=0;
			for(a=0;a<point_mcs.length;a++){
				set_wind_motion(a);
			}
		}
		public function set_wind_motion(no:int):void{
			var mc=wind_point_status[no];
			if(!mc.vector_flg)return;
			
			//======---set_sin---======
			mc.nowx=Math.sin(mc.brad)*(mc.brx)+mc.basebrx+mc.defx;
			mc.nowy=Math.sin(mc.brad)*(mc.bry)+mc.basebry+mc.defy;
			mc.nowz=Math.sin(mc.brad)*(mc.brz)+mc.basebrz+mc.defz;
			
			mc.brx*=0.985;
			mc.bry*=0.985;
			mc.brz*=0.985;
			mc.basebrx*=0.975;
			mc.basebry*=0.975;
			mc.basebrz*=0.975;
			mc.brad+=get_k2r()*3.7;
			if(mc.brad>Math.PI*2)mc.brad-=Math.PI*2;
			
			mc.basebrx=(Math.abs(mc.basebrx)<0.5)?0:mc.basebrx;
			mc.basebry=(Math.abs(mc.basebry)<0.5)?0:mc.basebry;
			mc.basebrz=(Math.abs(mc.basebrz)<0.5)?0:mc.basebrz;
			
			mc.brx=(Math.abs(mc.brx)<0.5)?0:mc.brx;
			mc.bry=(Math.abs(mc.bry)<0.5)?0:mc.bry;
			mc.brz=(Math.abs(mc.brz)<0.5)?0:mc.brz;
			if(mc.brx==0 && mc.bry==0 && mc.brz==0 && mc.basebrx==0 && mc.basebry==0 && mc.basebrz==0){
				mc.basebrx=0;
				mc.basebry=0;
				mc.basebrz=0;
				mc.brad=0;
				mc.nowx=mc.defx+0;
				mc.nowy=mc.defy+0;
				mc.nowz=mc.defz+0;
				mc.vector_flg=false;
			}
			//=======================
			
			//======---set_pos---======
			var a:int=0;
			var p=point_mcs[no];
			p.x=mc.nowx;
			p.y=mc.nowy;
			p.z=mc.nowz;
			//=======================
		}								
	
	
	
		public function get_3p(){
			var p=[{x:0,y:0}, {x:0,y:0}, {x:0,y:0}];
			p[0].x=this.geometry.faces[0].v0.vertex3DInstance.x;
			p[0].y=this.geometry.faces[0].v0.vertex3DInstance.y;
			p[1].x=this.geometry.faces[0].v1.vertex3DInstance.x;
			p[1].y=this.geometry.faces[0].v1.vertex3DInstance.y;
			p[2].x=this.geometry.faces[0].v2.vertex3DInstance.x;
			p[2].y=this.geometry.faces[0].v2.vertex3DInstance.y;
			return p;
		}
		
		public function get_face_points():void{
			var a:int=0;
			var m:Array;
			var o:Object;
			for(a=0;a<this.geometry.faces.length;a++){
				o=new Object();
				//this.geometry.faces[a].v0.vertex3DInstance.x;
			}
		}
		
		public function get_point_zsort_list(camera :CameraObject3D):Array{
			var a:int=0;
			var mc=pointer;
			var m:Array=[];
			var gap:Number=0;
			for(a=0;a<point_mcs.length;a++){
				mc.x=point_mcs[a].x;
				mc.y=point_mcs[a].y;
				mc.z=point_mcs[a].z;
				project_pointer(this, camera);
				//mc.world.calculateMultiply( this.world, this.transform );
				//mc.calculateScreenCoords(camera);
				gap=get_gap(mc, camera);
				//trace(mc.screen.x+" "+mc.screen.y);
				m[a]={no:a, gap:gap, p:point_mcs[a], screenX:mc.screen.x, screenY:mc.screen.y, screenZ:mc.screen.z};
			}
			m.sortOn("no", 16);
			return m;
		}
		
		public function project_pointer( parent :DisplayObject3D, camera):void{
			updateTransform();
			pointer.world.calculateMultiply( parent.world, pointer.transform);
			
			if( parent !== camera ){
				if(camera.useProjectionMatrix)
					pointer.view.calculateMultiply4x4( parent.view, pointer.transform );
				else
					pointer.view.calculateMultiply( parent.view,pointer.transform );
			}else{
				if(camera.useProjectionMatrix)
					pointer.view.calculateMultiply4x4(camera.eye, pointer.transform );
				else
					pointer.view.calculateMultiply(camera.eye, pointer.transform );
			}
			pointer.calculateScreenCoords(camera);
		}
		
		public function init_close(radius:Number):void{
			var a:int=0;
			var b:int=0;
			var r:Number=radius;
			var gap:Number=0;
			var gx:Number=0;
			var gy:Number=0;
			var gz:Number=0;
			for(a=0;a<point_mcs.length;a++){
				close_point_mcs[a]=new Array();
				point_exe_flgs[a]=false;
				for(b=0;b<point_mcs.length;b++){
					if(a==b)continue;
					gx=point_mcs[a].x-point_mcs[b].x;
					gy=point_mcs[a].y-point_mcs[b].y;
					gz=point_mcs[a].z-point_mcs[b].z;
					gap=Math.sqrt( gx*gx + gy*gy + gz*gz );
					//if(b==0)trace(gap);
					if(gap<=r)close_point_mcs[a].push(point_mcs[b]);
				}
				//trace(close_point_mcs[a].length);
			}
		}
		
		public function get_gap(a, b):Number{
			var gx:Number=a.x-b.x;
			var gy:Number=a.y-b.y;
			var gz:Number=a.z-b.z;
			return Math.sqrt( gx*gx + gy*gy + gz*gz );
		}
			
		
		public function calculate_vector():void{
			var a:int=0;
			var mc;
			var pm:Number=0;
			var pm2:Number=0;
			for(a=0;a<point_mcs.length;a++){
				mc=point_mcs[a];
				point_vector_mcs[a]+=point_gravity_mcs[a]*((mc.z>0)?-1:1);
				
				pm=(mc.z>0)?1:-1;
				pm2=((mc.z+point_vector_mcs[a])>0)?1:-1;
				if(pm!=pm2)point_gravity_mcs[a]*=0.5;
			}
		}
		
		public function add_vector():void{
			var a:int=0;
			var mc;
			for(a=0;a<point_mcs.length;a++){
				mc=point_mcs[a];
				mc.z+=point_vector_mcs[a];
			}
		}
		
		private function buildPlane( width:Number, height:Number ):void
		{
			var gridX    :Number = this.segmentsW;
			var gridY    :Number = this.segmentsH;
			var gridX1   :Number = gridX + 1;
			var gridY1   :Number = gridY + 1;
	
			var vertices :Array  = this.geometry.vertices;
			var faces    :Array  = this.geometry.faces;
	
			var textureX :Number = width /2;
			var textureY :Number = height /2;
	
			var iW       :Number = width / gridX;
			var iH       :Number = height / gridY;
	
			var cnt:Number=0;
			
			// Vertices
			for( var ix:int = 0; ix < gridX + 1; ix++ ){
				point_list[ix]=new Array();
				for( var iy:int = 0; iy < gridY1; iy++ ){
					var x :Number = ix * iW - textureX;
					var y :Number = iy * iH - textureY;
	
					vertices.push( new Vertex3D( x, y, 0 ) );
					
					point_list[ix][gridY1-iy-1]=vertices[vertices.length-1];
					point_mcs[point_mcs.length]=vertices[vertices.length-1];
					point_vector_mcs[point_vector_mcs.length]=0;
					point_gravity_mcs[point_gravity_mcs.length]=0;
					vertices[vertices.length-1].extra=new Object();
					vertices[vertices.length-1].extra.no=cnt;
					cnt++;
				}
			}
	
			var m:int=0;
			var n:int=0;
			for(m=0;m<point_list.length;m++){
				for(n=0;n<point_list[m].length;n++){
					if(point_list2[n]==undefined || point_list2[n]==null)point_list2[n]=new Array();
					point_list2[n][m]=point_list[m][n];
				}
			}
			
			
			// Faces
			var uvA :NumberUV;
			var uvC :NumberUV;
			var uvB :NumberUV;
	
			for(  ix = 0; ix < gridX; ix++ )
			{
				for(  iy= 0; iy < gridY; iy++ )
				{
					// Triangle A
					var a:Vertex3D = vertices[ ix     * gridY1 + iy     ];
					var c:Vertex3D = vertices[ ix     * gridY1 + (iy+1) ];
					var b:Vertex3D = vertices[ (ix+1) * gridY1 + iy     ];
	
					uvA =  new NumberUV( ix     / gridX, iy     / gridY );
					uvC =  new NumberUV( ix     / gridX, (iy+1) / gridY );
					uvB =  new NumberUV( (ix+1) / gridX, iy     / gridY );
	
					faces.push(new Triangle3D(this, [ a, b, c ], material, [ uvA, uvB, uvC ] ) );
	
					// Triangle B
					a = vertices[ (ix+1) * gridY1 + (iy+1) ];
					c = vertices[ (ix+1) * gridY1 + iy     ];
					b = vertices[ ix     * gridY1 + (iy+1) ];
	
					uvA =  new NumberUV( (ix+1) / gridX, (iy+1) / gridY );
					uvC =  new NumberUV( (ix+1) / gridX, iy      / gridY );
					uvB =  new NumberUV( ix      / gridX, (iy+1) / gridY );
					
					faces.push(new Triangle3D(this, [ a, b, c ], material, [ uvA, uvB, uvC ] ) );
				}
			}
	
			this.geometry.ready = true;
			
			if(Papervision3D.useRIGHTHANDED)
				this.geometry.flipFaces();
		}
	}
}